// Copyright (C) 2020 Intel Corporation

// SPDX-License-Identifier: MIT

#include <CL/sycl.hpp>
#include <chrono>
using namespace sycl;

// Appropriate values depend on your HW
constexpr int BLOCK_SIZE = 42;
constexpr int NUM_BLOCKS = 2500;
constexpr int N = NUM_BLOCKS * BLOCK_SIZE;
constexpr int TEST_LOOP = 10;
constexpr int GUESS_PAGE = 1024;

int main() {
  //queue Q{cpu_selector{}};
  queue Q;
  int *data = malloc_shared<int>(N, Q);
  int *read_only_data = malloc_shared<int>(BLOCK_SIZE, Q);

  // Never updated after initialization
  for (int i = 0; i < BLOCK_SIZE; i++)
    read_only_data[i] = i;
  for (int i = 0; i < N; i++)
    data[i] = 0; // done on the host side
  // Mark this data as "read only" so the runtime can copy it
  // to the device instead of migrating it from the host.
  // Real values will be documented by your DPC++ backend.
  int HW_SPECIFIC_ADVICE_RO = 0;
    using ns = std::chrono::nanoseconds;

 
  Q.mem_advise(read_only_data, BLOCK_SIZE,(pi_mem_advice)HW_SPECIFIC_ADVICE_RO);
  auto start = std::chrono::steady_clock::now();
 //event e = Q.prefetch(data, BLOCK_SIZE);
 // e.wait();

  // for (int b = 1; b < NUM_BLOCKS; b++) {
  //       if ((b + 1) < NUM_BLOCKS) {
  //         // Prefetch next block
  //         e = Q.prefetch(data + (b + 1) * BLOCK_SIZE, 1);//BLOCK_SIZE);
  //         e.wait();
  //       }
  // }

   // Q.parallel_for(range{N},//e,
    //               [=](id<1> i) { data[i] =0; });  
   
    //Q.wait();
  //  for (int b = 0; b < NUM_BLOCKS; b++) {
  //   Q.parallel_for(range{1},//e,
  //                  [=](id<1> i) { data[b * BLOCK_SIZE + i] =0; });  
  //   }
  //   Q.wait();

   for (int b = 0; b <  N; b+=GUESS_PAGE) {
    Q.single_task([=](){
      // only need to read first data of page
      // to get the page into device memory
                  int x =data[b]; });  
                  //data[b] =0; });  
    }
    //Q.wait();



  //  for (int b = 0; b < NUM_BLOCKS; b++) {
  //   Q.parallel_for(range{1},//e,
  //                  [=](id<1> i) { data[b * BLOCK_SIZE + i] =0; });  
  //   }
  //   Q.wait();
  auto duration = std::chrono::steady_clock::now() - start;
  auto time = std::chrono::duration_cast<ns>(duration).count();
  std::cout << "setup-time:" << time/1.0E9 << "\n";

  // e = Q.prefetch(data, BLOCK_SIZE);
  for (int j=0;j<TEST_LOOP;j++){
  auto start = std::chrono::steady_clock::now();

  for (int b = 0; b < NUM_BLOCKS; b++) {
    Q.parallel_for(range{BLOCK_SIZE},//e,
                   [=](id<1> i) { data[b * BLOCK_SIZE + i] += read_only_data[i]; });

    
    // if ((b + 1) < NUM_BLOCKS) {
    //   // Prefetch next block
    //   e = Q.prefetch(data + (b + 1) * BLOCK_SIZE, BLOCK_SIZE);
    // }
  }
  Q.wait();
  auto duration = std::chrono::steady_clock::now() - start;
  auto time = std::chrono::duration_cast<ns>(duration).count();
  std::cout << "time:" << time/1.0E9 << "\n";

    }

  bool passed = true;
  for (int i=0;i<N;i++){
    int x = data[i];
    int ii = i%BLOCK_SIZE;
    int j= x/TEST_LOOP;
    passed = passed && j==ii;
  }

  std::cout << (passed? "pass" : "fail") <<"\n";

  free(data, Q);
  free(read_only_data, Q);
  return 0;
}
